/*
  Extension       JsonValidator
  Author          Carlo Hogeveen. Syntax validator by David Goodenough.
  Website         eCarlo.nl/tse
  Compatibility   TSE v4 upwards, all variants
  Version         v0.2   BETA   31 Aug 2024


  This extension validates JSON syntax for a JSON buffer.

  The current buffer is validated in the following cases:
  - When Validate is selected from its menu.
  - When executed with the "validate" parameter.
  - When a .json file is saved.
  - When a .json is edited.

  An "error cursor" is shown if the error location is in the editing window.

  The "error location" is where the syntax checker notices illegal syntax.
  The actual error can be a little bit before that.

  There is a configuration menu to tweak some of the above functionality.


  INSTALLATION

  Copy this file to TSE's "mac" directory and compile it there, for example by
  opening it there with TSE and applying the Macro -> Compile menu.

  Add this macro's name ("JsonValidator") to TSE's Macro AutoLoad List.

  Execute the macro to configure it.

  Note, by configuring JsonValidator to NOT save a .json file with errors,
  you can safely use TSE's own "File Save" key to validate such a file.

  Optionally define TSE keys
    For starting the menu
      <a key name>  ExecMacro('JsonValidator')
    Or to validate the current buffer
      <a key name>  ExecMacro('JsonValidator validate')



  HISTORY

  v0.0.0.1   DEVELOPMENT   27 Aug 2024
    Initial version.

  v0.1       BETA          28 Aug 2024
    Added solutions for deficiencies.
    Fixed bugs.

  v0.2       BETA          31 Aug 2024
    Simplified the documentation.

    Fixed the color picker's header not being shown.

    Created 7 configuration options for saving a .json file that has errors,
    including options for NOT saving such a file.
    Therefore some options allow using TSE's own "File Save" key, for example
    <Ctrl S>, to validate a .json file without saving it if it has errors.

*/





// Start of compatibility restrictions and mitigations ...


#ifndef INTERNAL_VERSION
  #define INTERNAL_VERSION 0
#endif


#if INTERNAL_VERSION == 0
  string proc GetTempPath()
    string result [MAXSTRINGLEN] = ''
    #ifdef LINUX
      result = '/tmp/'
    #else
      result = GetEnvStr('TMP')
      if result == ''
        result = GetEnvStr('TEMP')
      endif
      if result <> ''
        result = AddTrailingSlash(result)
      endif
    #endif
    return(result)
  end GetTempPath
#endif


#if EDITOR_VERSION < 4400h
  /*
    StrReplace() 1.0

    If you have TSE Pro 4.0 or 4.2, then this proc almost completely implements
    the built-in StrReplace() function of TSE Pro 4.4.
    The StrReplace() function replaces a string (pattern) inside a string.
    It works for strings like the Replace() function does for files, so read
    the Help for the Replace() function for the usage of the options, but apply
    these differences while reading:
    - Where Replace() refers to "file" and "line", StrReplace() refers to
      "string".
    - The options "g" ("global", meaning "from the start of the string")
      and "n" ("no questions", meaning "do not ask for confirmation on
      replacements") are implicitly always active, and can therefore be omitted.
    Notable differences between the procedure below with TSE 4.4's built-in
    function are, that here the fourth parameter "options" is mandatory
    and that the fifth parameter "start position" does not exist.
  */
  integer strreplace_id = 0
  string proc StrReplace(string needle, string haystack, string replacer, string options)
    integer i                      = 0
    integer org_id                 = GetBufferId()
    string  result  [MAXSTRINGLEN] = haystack
    string  validated_options [20] = 'gn'
    for i = 1 to Length(options)
      if (Lower(SubStr(options, i, 1)) in '0'..'9', 'b', 'i','w', 'x', '^', '$')
        validated_options = validated_options + SubStr(options, i, 1)
      endif
    endfor
    if strreplace_id == 0
      strreplace_id = CreateTempBuffer()
    else
      GotoBufferId(strreplace_id)
      EmptyBuffer()
    endif
    InsertText(haystack, _INSERT_)
    lReplace(needle, replacer, validated_options)
    result = GetText(1, CurrLineLen())
    GotoBufferId(org_id)
    return(result)
  end StrReplace
#endif


// End of compatibility restrictions and mitigations.





/*
  This extension includes v2 of David Goodenough's json.s macro.
  That macro checks whether the current buffer contains valid JSON syntax.

  I made 4 tiny changes to the included version:
  - The explicit key definition is not included.
    Alternatives:
      Saving a .json file when JsonValidator is loaded.
      Optionally assigned to a TSE key:
        ExecMacro('JsonValidator validate')
        ExecMacro('JsonValidator') and select "Validate" from the menu.
  - The Message()s are now surrounded by an
      if Query(MsgLevel) == _ALL_MESSAGES_
        ...
      endif
    so that they can be hidden when necessary.
  - ParseJson() is no longer a public proc.
  - ParseJson() now returns TRUE or FALSE for whether errors were found.

*/


// Start of David Goodenough's json.s v2 code ...


forward integer proc ParseElement(integer name)

#define Tab             9
#define SPACE           32
#define DOUBLE_QUOTE    34
#define PLUS            43
#define COMMA           44
#define MINUS           45
#define PERIOD          46
#define ZERO            48
#define NINE            57
#define COLON           58
#define UPPER_E         69
#define LOWER_E         101
#define BACKSLASH       92
#define OPEN_SQUARE     91
#define CLOSE_SQUARE    93
#define OPEN_BRACE      123
#define CLOSE_BRACE     125

// Get the next non-white space character.  Return -1 for end of
// buffer, otherwise step over tabs spaces and line ends.
integer proc GetNextChar()
    integer ch = 0
    repeat
        ch = CurrChar()
        if ch in SPACE, Tab, _AT_EOL_, _BEYOND_EOL_
            if not NextChar()
                return (-1)
            endif
        endif
    until not (ch in SPACE, Tab, _AT_EOL_, _BEYOND_EOL_)
    return (ch)
end

// Try and match the string token passed.  If it is found, advance
// the cursor past it and return succcess, otherwise leave the cursor
// unchanged and return failure.
integer proc ParseToken(string token)
    integer len = Length(token)
    integer leftOnLine = CurrLineLen() - CurrPos() + 1
    string test[8]
    if leftOnLine < len
        return (0)
    endif
    test = GetText(CurrPos(), len)
    if test <> token
        return (0)
    endif
    GotoPos(CurrPos() + len)
    return (1)
end

// Parse a quoted string.  It just consumes characters until it finds the
// closing quote, ignoring characters that follow a backslash.  If it falls
// off the end of the line without seeing the quote, it returns an error.
integer proc ParseString()
    integer ch = GetNextChar()
    if ch == -1
        return (-1)
    endif
    if ch <> DOUBLE_QUOTE
        return (0)
    endif

    repeat
        Right()
        if CurrPos() > CurrLineLen()
            return (0)
        endif
        ch = CurrChar()
        if ch == DOUBLE_QUOTE
            NextChar()
            return (1)
        endif
        if ch == BACKSLASH
            Right()
            if CurrPos() > CurrLineLen()
                return (0)
            endif
        endif
    until 0
    return (0)
end

// Parse the name for a value in an object.  This just looks for a string followed
// by a colon.
integer proc ParseName()
    integer ok = ParseString()
    integer ch = 0
    if ok <> 1
        return (ok)
    endif

    ch = GetNextChar()
    if ch == -1
        return (-1)
    endif
    if ch <> COLON
        return (0)
    endif

    NextChar()
    return (1)
end

// Parse a JSON value.  This first checks for the three tokens: "null",
// "true" and "false".  It then looks for a leading digit or minus sign
// and if found it parses a number/  Otherwise it checkes for a double
// quote which causes it to parse a string.  If nothing matches return
// an error condition.
integer proc ParseValue()
    integer ch = GetNextChar()
    integer ok = 0
    if ch == -1
        return (-1)
    endif

    if ParseToken("null")
        return (1)
    endif

    if ParseToken("true")
        return (1)
    endif

    if ParseToken("false")
        return (1)
    endif

    // Number parsing is a bit janky.  If we see
    // a leading digit or minus sign, assume that
    // any sequence of digits, plus or minus signs
    // periods or letter 'e' in either case form
    // a valid number.
    // This will let a few errors slip through, but
    // I am simply not motivated enough to write a
    // full blown numeric parser for this.
    if (ch >= ZERO and ch <= NINE) or ch == MINUS
        repeat
            Right()
            if CurrPos() > CurrLineLen()
                ch = SPACE
            else
                ch = CurrChar()
            endif
        until (ch < ZERO or ch > NINE) and not (ch in PLUS, MINUS, PERIOD, UPPER_E, LOWER_E)

        return (1)
    endif

    if ch == DOUBLE_QUOTE
        ok = ParseString()
        return (ok)
    endif

    return (0)
end

// Parse an array.  On entry, the open square has been seen, so the first
// thing done is to look for a close square which is a valid JSON empty array.
// If that is not found, repeatedlu parse values, looking for a comma to
// separate or a close square to terminate the array
integer proc ParseArray()
    integer ch = 0
    integer ok = 0
    //string result[128]
    if not NextChar()
        return (-1)
    endif

    ch = GetNextChar()
    if ch == -1
        return (-1)
    endif
    if ch == CLOSE_SQUARE
        //Warn("ParseArray empty array, early out")
        NextChar()
        return (1)
    endif

    repeat
        ok = ParseElement(FALSE)
        //result = "ParseElement in ParseArray " + format(ok:3)
        //Warn(result)
        if ok <> 1
            return (ok)
        endif
        ch = GetNextChar()
        //result = "GetNextChar in ParseArray " + format(ch:3)
        //Warn(result)
        if ch == -1
            return (-1)
        endif

        if ch == CLOSE_SQUARE
            //Warn("ParseArray saw ]")
            NextChar()
            return (1)
        endif
        if ch <> COMMA
            //Warn("ParseArray did not see ,")
            return (0)
        endif
        if not NextChar()
            return (-1)
        endif
        //Warn("ParseArray iterating")
    until 0
    return (0)
end

// Parse an object.  This uses the exact same logic as ParseArray() above
// except that when parsing elements, it passes a true value to ParseElement()
// to cause it to parse the name as well as the value.
integer proc ParseObject()
    integer ch = 0
    integer ok = 0
    //string result[128]
    if not NextChar()
        return (-1)
    endif

    ch = GetNextChar()
    if ch == -1
        return (-1)
    endif
    if ch == CLOSE_BRACE
        //Warn("ParseObject empty object, early out")
        NextChar()
        return (1)
    endif

    repeat
        ok = ParseElement(TRUE)
        //result = "ParseElement in ParseObject " + format(ok:3)
        //Warn(result)
        if ok <> 1
            return (ok)
        endif
        ch = GetNextChar()
        //result = "GetNextChar in ParseObject " + format(ch:3)
        //Warn(result)
        if ch == -1
            return (-1)
        endif
        if ch == CLOSE_BRACE
            //Warn("ParseObject saw }")
            NextChar()
            return (1)
        endif
        if ch <> COMMA
            //Warn("ParseObject did not see ,")
            return (0)
        endif
        if not NextChar()
            return (-1)
        endif
        //Warn("ParseObject iterating")
    until 0
    return (0)
end

// Parse a single element, typically a member of either an object or an
// array.  If we are parsing an object, name will be true, and it checks
// for a leading name.  If falsse, it just looks for a single value
integer proc ParseElement(integer name)
    integer ok = 0
    integer ch = 0
    if name
        ok = ParseName()
        if ok <> 1
            return (ok)
        endif
    endif

    ch = GetNextChar()
    if ch == -1
        return (-1)
    endif

    if ch == OPEN_SQUARE
        ok = ParseArray()
    elseif ch == OPEN_BRACE
        ok = ParseObject()
    else
        ok = ParseValue()
    endif

    return (ok)
end

// Top level parser.  Due to the behavior of ParsElement() there is a bug here where
// a single JSON value, while valid, is not correctly parsed.  Fixing this would take
// a rewrite of the top of ParseElement() that I Am not sure I want to take on.
integer proc ParseJSON()
    integer ok = 0
    integer ch = GetNextChar()

    if ch == DOUBLE_QUOTE
        ok = ParseElement(TRUE)
    elseif ch == OPEN_SQUARE
        ok = ParseArray()
    elseif ch == OPEN_BRACE
        ok = ParseObject()
    endif

    if Query(MsgLevel) == _ALL_MESSAGES_
        if ok == 1
            Message("Valid JSON")
        elseif ok == 0
            Message("JSON Syntax error")
        else
            Message("EOF reached: JSON incomplete")
        endif
    endif
    return(ok == 1)
end


// End of David Goodenough's json.s v2 code.





// Global constants and pseudo-constants
#define CCF_OPTIONS                           _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_
string  CFG_SECTION_NAME     [MAXSTRINGLEN] = ''
string  ERROR_LINE_VAR_NAME  [MAXSTRINGLEN] = ''
string  ERROR_POS_VAR_NAME   [MAXSTRINGLEN] = ''
string  FILECHANGED_VAR_NAME [MAXSTRINGLEN] = ''
string  MACRO_NAME           [MAXSTRINGLEN] = ''
#define MAX_COLOR_NAME_SIZE                   33
string  RECHECK_VAR_NAME     [MAXSTRINGLEN] = ''
string  SAVE_OPTIONS                  [116] = 'Ask, Default No;Ask, Default Yes;No, with temporary message;Yes, with temporary message;No, silently;Yes, silently'
string  TMP_FQN              [MAXSTRINGLEN] = ''



// Global variables
integer cfg_error_attr         = 0
integer cfg_save_option        = 0
integer cfg_message_duration   = 0
integer idle_timer             = 0
string  old_fqn [MAXSTRINGLEN] = ''
integer old_pos_row            = 0
integer old_pos_xoffset        = 0
integer old_undocount          = 0
integer old_undomode           = 0
integer prev_error_state       = FALSE
integer restart_main_menu      = FALSE
integer restore_id             = 0
string  saved_attr         [1] = ''


proc to_beep_or_not_to_beep()
  if Query(Beep)
    Alarm()
  endif
end to_beep_or_not_to_beep


integer proc write_profile_int(string  section_name,
                               string  item_name,
                               integer item_value)
  integer result = WriteProfileInt(section_name,
                                   item_name,
                                   item_value)
  if not result
    to_beep_or_not_to_beep()
    Warn('ERROR writing configuration to file "tse.ini".')
  endif
  return(result)
end write_profile_int


proc push_position()
  PushLocation()
  old_pos_xoffset = CurrXoffset()
  old_pos_row     = CurrRow()
end push_position


proc pop_position()
  PopLocation()

  if     CurrXoffset() < old_pos_xoffset
    RollRight(old_pos_xoffset - CurrXoffset())
  elseif CurrXoffset() > old_pos_xoffset
    RollLeft(CurrXoffset() - old_pos_xoffset)
  endif

  if CurrRow() < old_pos_row
    RollDown(old_pos_row - CurrRow())
  elseif CurrRow() > old_pos_row
    RollUp(CurrRow() - old_pos_row)
  endif
end pop_position


proc pop_msg(string window_title, string msg, integer duration)
  integer left_margin  = 1
  integer popup_width  = Max(Length(window_title), Length(msg)) + 2
  integer right_margin = 0

  if duration > 0
    if Length(msg) + 2 < Query(ScreenCols)
      left_margin = Max((Query(ScreenCols) - popup_width) / 2, 1)
    endif

    right_margin = left_margin + popup_width - 1

    PopWinOpen(left_margin,
               Query(ScreenRows) / 2 - 1,
               right_margin,
               Query(ScreenRows) / 2 + 1,
               1, window_title, Query(MenuBorderAttr))
    ClrScr()
    PutStrAttrXY(1, 1, msg, '', Query(MenuTextAttr))
    Delay(duration)
    PopWinClose()
  endif
end pop_msg


proc pop_message(string msg)
  pop_msg(MACRO_NAME, msg, cfg_message_duration * 18)
end pop_message


proc show_help()
  string  full_macro_source_name [MAXSTRINGLEN] = SplitPath(CurrMacroFilename(), _DRIVE_|_PATH_|_NAME_) + '.s'
  string  help_file_name         [MAXSTRINGLEN] = '*** ' + MACRO_NAME + ' Help ***'
  integer hlp_id                                = GetBufferId(help_file_name)
  integer org_id                                = GetBufferId()
  integer tmp_id                                = 0

  if hlp_id
    GotoBufferId(hlp_id)
    UpdateDisplay()
  else
    tmp_id = CreateTempBuffer()
    if LoadBuffer(full_macro_source_name)
      // Separate characters, otherwise my SynCase macro gets confused.
      if lFind('/' + '*', 'g')
        PushBlock()
        UnMarkBlock()
        Right(2)
        MarkChar()
        if not lFind('*' + '/', '')
          EndFile()
        endif
        MarkChar()
        Copy()
        CreateTempBuffer()
        Paste()
        UnMarkBlock()
        PopBlock()
        BegFile()
        ChangeCurrFilename(help_file_name, CCF_OPTIONS)
        BufferType(_NORMAL_)
        FileChanged(FALSE)
        BrowseMode(TRUE)
        UpdateDisplay()
      else
        GotoBufferId(org_id)
        Warn('File "', full_macro_source_name, '" has no multi-line comment block.')
      endif
    else
      GotoBufferId(org_id)
      Warn('File "', full_macro_source_name, '" not found.')
    endif
    AbandonFile(tmp_id)
  endif
end show_help


string proc get_color_name(integer color_attr)
  integer color_background                 = color_attr  /  16
  integer color_foreground                 = color_attr mod 16
  string  color_name [MAX_COLOR_NAME_SIZE] = 'Intense Bright Magenta on Magenta'
  string  color_names                 [46] = 'Black Blue Green Cyan Red Magenta Yellow White'

  color_name = iif(color_foreground > 7, 'Bright ' , '')
             + GetToken(color_names, ' ', (color_foreground mod 8) + 1)
             + ' on '
             + iif(color_background > 7, 'Intense ', '')
             + GetToken(color_names, ' ', (color_background mod 8) + 1)
  return(color_name)
end get_color_name


string proc get_color_usages(integer color_attr)
  string result [MAXSTRINGLEN] = ''
  if color_attr == Query(TextAttr)
    result = result + ', Text'
  endif
  if color_attr == Query(HiliteAttr)
    result = result + ', Hilite'
  endif
  if color_attr == Query(BlockAttr)
    result = result + ', Block'
  endif
  if color_attr == Query(CursorAttr)
    result = result + ', Cursor'
  endif
  if color_attr == Query(CursorInBlockAttr)
    result = result + ', CursorInBlock'
  endif
  if color_attr == Query(Directive1Attr)
    result = result + ', Directive1'
  endif
  if color_attr == Query(Directive2Attr)
    result = result + ', Directive2'
  endif
  if color_attr == Query(Directive3Attr)
    result = result + ', Directive3'
  endif
  if color_attr == Query(MultiLnDlmt1Attr)
    result = result + ', MultiLnDlmt1'
  endif
  if color_attr == Query(MultiLnDlmt2Attr)
    result = result + ', MultiLnDlmt2'
  endif
  if color_attr == Query(MultiLnDlmt3Attr)
    result = result + ', MultiLnDlmt3'
  endif
  if color_attr == Query(SingleLnDlmt1Attr)
    result = result + ', SingleLnDlmt1'
  endif
  if color_attr == Query(SingleLnDlmt2Attr)
    result = result + ', SingleLnDlmt2'
  endif
  if color_attr == Query(SingleLnDlmt3Attr)
    result = result + ', SingleLnDlmt3'
  endif
  if color_attr == Query(ToEol1Attr)
    result = result + ', ToEol1'
  endif
  if color_attr == Query(ToEol2Attr)
    result = result + ', ToEol2'
  endif
  if color_attr == Query(ToEol3Attr)
    result = result + ', ToEol3'
  endif
  if color_attr == Query(Quote1Attr)
    result = result + ', Quote1'
  endif
  if color_attr == Query(Quote2Attr)
    result = result + ', Quote2'
  endif
  if color_attr == Query(Quote3Attr)
    result = result + ', Quote3'
  endif
  if color_attr == Query(NumberAttr)
    result = result + ', Number'
  endif
  if color_attr == Query(IncompleteQuoteAttr)
    result = result + ', IncompleteQuote'
  endif
  if color_attr == Query(Keywords1Attr)
    result = result + ', Keywords1'
  endif
  if color_attr == Query(Keywords2Attr)
    result = result + ', Keywords2'
  endif
  if color_attr == Query(Keywords3Attr)
    result = result + ', Keywords3'
  endif
  if color_attr == Query(Keywords4Attr)
    result = result + ', Keywords4'
  endif
  if color_attr == Query(Keywords5Attr)
    result = result + ', Keywords5'
  endif
  if color_attr == Query(Keywords6Attr)
    result = result + ', Keywords6'
  endif
  if color_attr == Query(Keywords7Attr)
    result = result + ', Keywords7'
  endif
  if color_attr == Query(Keywords8Attr)
    result = result + ', Keywords8'
  endif
  if color_attr == Query(Keywords9Attr)
    result = result + ', Keywords9'
  endif
  result = SubStr(result, 3, MAXSTRINGLEN)
  result = StrReplace(', ', result, ' and ', 'b1')
  return(result)
end get_color_usages


integer proc is_valid_text_color(integer color_attr)
  integer result           = TRUE
  integer background_color = color_attr  /  16
  integer foreground_color = color_attr mod 16
  if foreground_color == background_color
    result = FALSE
  endif
  return(result)
end is_valid_text_color


integer proc select_color(integer param_color, string param_prompt)
  integer background                  = 0
  string  color_name             [33] = 'Intense Bright Magenta on Magenta'
  integer cols_per_color              = (Query(ScreenCols) - 2) / 16
  integer foreground                  = 0
  integer grid_color                  = 0
  integer left_margin                 = 0
  integer mouse_background            = 0
  integer mouse_foreground            = 0
  integer old_attr                    = Set(Attr, Color(Bright White ON Black))
  integer old_cursor                  = Set(Cursor, OFF)
  integer old_SpecialEffects          = 0
  integer prev_clockticks             = 0
  integer rows_per_color              = (Query(ScreenRows) - 3) / 16
  integer selected_color              = param_color
  integer selected_background         = 0
  integer selected_foreground         = 0
  integer sub_row                     = 0
  integer top_margin                  = (Query(ScreenRows) - 3 - rows_per_color * 16) / 2 + 1
  string  window_title [MAXSTRINGLEN] = ''

          left_margin                 = (Query(ScreenCols) - 2 - cols_per_color * 16) / 2 + 1
  BufferVideo()
  ClrScr()

  if param_color <> 0
     selected_foreground = param_color mod 16
     selected_background = param_color  /  16
  endif

  repeat
    color_name = get_color_name(selected_background * 16 + selected_foreground)

    if param_prompt == ''
      window_title = color_name
    else
      window_title = Format(param_prompt, color_name:36)
    endif

    PopWinOpen(left_margin,
               top_margin,
               left_margin + (cols_per_color * 16) + 1,
               top_margin  + (rows_per_color * 16) + 1,
               1,
               window_title,
               Color(Bright White ON Black))

    for foreground = 0 to 15 // Rows.
      for background = 0 to 15 // Columns.
        for sub_row = 1 to rows_per_color
          if foreground <> 0 // Skip a bug in the PutStrAttrXY command
          or background <> 0 // up to and including TSE 4.2.
            grid_color = background * 16 + foreground
            PutStrAttrXY(background * cols_per_color + 1,
                         foreground * rows_per_color + sub_row,
                         Format((grid_color)
                                :cols_per_color:'0':16),
                         '',
                         grid_color)
          endif
        endfor
      endfor
    endfor

    old_SpecialEffects = Set(SpecialEffects,
                         Query(SpecialEffects) & ~ _DRAW_SHADOWS_)
    PopWinOpen(left_margin + selected_background * cols_per_color,
               top_margin  + selected_foreground * rows_per_color,
               left_margin + selected_background * cols_per_color + cols_per_color + 1,
               top_margin  + selected_foreground * rows_per_color + rows_per_color + 1,
               1,
               '',
               Color(Bright White ON Black))
    Set(SpecialEffects, old_SpecialEffects)
    UnBufferVideo()

    case GetKey()
      when <Home>, <GreyHome>
        selected_foreground = 0
        selected_background = 0
      when <end>, <GreyEnd>
        selected_foreground = 15
        selected_background = 15
      when <CursorDown>, <GreyCursorDown>
        if selected_foreground == 15
          selected_foreground = 0
        else
          selected_foreground = selected_foreground + 1
        endif
      when <CursorUp>, <GreyCursorUp>
        if selected_foreground == 0
          selected_foreground = 15
        else
          selected_foreground = selected_foreground - 1
        endif
      when <CursorRight>, <GreyCursorRight>
        if selected_background == 15
          selected_background = 0
        else
          selected_background = selected_background + 1
        endif
      when <CursorLeft>, <GreyCursorLeft>
        if selected_background == 0
          selected_background = 15
        else
          selected_background = selected_background - 1
        endif
      when <LeftBtn>
        /*
          The Delay(1) turned out to be necessary to give the system
          time to update the mouse status. An extra call to MouseStatus
          was useless, as could be expected, because TSE should already
          do that internally after a mousekey.
        */
        Delay(1)
        mouse_background = (Query(MouseX) - left_margin - 1) / cols_per_color
        mouse_foreground = (Query(MouseY) -  top_margin - 1) / rows_per_color
        if  (mouse_background in 0 .. 15)
        and (mouse_foreground in 0 .. 15)
          if  mouse_background == selected_background
          and mouse_foreground == selected_foreground
            if GetClockTicks() - prev_clockticks < 18
              PushKey(<Enter>)
            endif
          else
            selected_background = mouse_background
            selected_foreground = mouse_foreground
          endif
          prev_clockticks = GetClockTicks()
        endif
    endcase

    BufferVideo()
    PopWinClose()
    PopWinClose()
  until Query(Key) in <Enter>, <GreyEnter>, <Escape>

  UpdateDisplay(_ALL_WINDOWS_REFRESH_)
  UnBufferVideo()
  Set(Attr  , old_attr  )
  Set(Cursor, old_cursor)

  if Query(Key) <> <Escape>
    selected_color = selected_background * 16 + selected_foreground
  endif

  return(selected_color)
end select_color


proc configure_color(string target)
  integer accept_selection   = FALSE
  integer stop_interrogation = FALSE
  integer new_attr           = 0
  integer old_attr           = cfg_error_attr
  integer response           = _NONE_

  new_attr = old_attr

  repeat
    new_attr = select_color(new_attr, 'Pick color for ' + target)
    if is_valid_text_color(new_attr)
      if new_attr == old_attr
        stop_interrogation = TRUE
      elseif get_color_usages(new_attr) == ''
        accept_selection   = TRUE
        stop_interrogation = TRUE
      else
        response = MsgBox('Are you sure?',
                          'The selected color "'
                          + get_color_name(new_attr)
                          + '" is already used for '
                          + get_color_usages(new_attr)
                          + '.',
                         _YES_NO_CANCEL_)
        accept_selection   = (response == 1)
        stop_interrogation = (response <> 2)
      endif
    else
      Warn('Do not make text invisible with the same color for foreground and background.')
    endif
  until stop_interrogation

  if     new_attr == old_attr
    Message('You did not change the color.')
  elseif accept_selection
    cfg_error_attr = new_attr
    write_profile_int(CFG_SECTION_NAME, 'ErrorColor', new_attr)
  endif
end configure_color


integer proc is_json_buffer()
  return(    BufferType() == _NORMAL_
         and BinaryMode() == 0
         and CurrExt() in '.json')
end is_json_buffer


integer proc check_for_json_syntax_errors(integer display_type)
  integer old_MsgLevel    = 0
  integer ok              = FALSE
  integer new_error_state = 0

  PushPosition()
  BegFile()

  if display_type == _HIDDEN_
    old_MsgLevel = Set(MsgLevel, _WARNINGS_ONLY_)
  endif

  ok = ParseJson()

  if display_type == _HIDDEN_
    Set(MsgLevel, old_MsgLevel)
  endif

  if ok
    new_error_state = FALSE

    SetBufferInt(ERROR_LINE_VAR_NAME, 0)
    SetBufferInt(ERROR_POS_VAR_NAME , 0)
    PopPosition()
  else
    new_error_state = TRUE

    SetBufferInt(ERROR_LINE_VAR_NAME, CurrLine())
    SetBufferInt(ERROR_POS_VAR_NAME , CurrPos() )

    if display_type == _HIDDEN_
      PopPosition()
    else
      KillPosition()
    endif
  endif

  if new_error_state <> prev_error_state
    if new_error_state
      Message('JSON syntax error found')
    else
      Message('No more JSON syntax errors')
    endif

    UpdateDisplay(_ALL_WINDOWS_REFRESH_)
    prev_error_state = new_error_state
  endif

  return(new_error_state)
end check_for_json_syntax_errors


proc show_error_cursor()
  integer error_line         = GetBufferInt(ERROR_LINE_VAR_NAME)
  integer error_pos          = GetBufferInt(ERROR_POS_VAR_NAME )
  string  error_pos_attr [1] = ''
  string  error_pos_char [1] = ''
  integer error_scr_col      = 0
  integer error_scr_row      = 0
  integer error_win_col      = 0
  integer error_win_row      = 0
  integer win_line_from      = CurrLine() - CurrRow() + 1
  integer win_line_to        = win_line_from + Query(WindowRows) - 1

  if CurrLine() in win_line_from .. win_line_to
    error_win_row = error_line - win_line_from + 1
    error_win_col = error_pos  - CurrXoffset()

    if error_win_col in 1 .. Query(WindowCols)
      error_scr_row = Query(WindowY1) + error_win_row - 1
      error_scr_col = Query(WindowX1) + error_win_col - 1
      if GetStrAttrXY(error_scr_col, error_scr_row,
                      error_pos_char, error_pos_attr, 1) == 1
        if Asc(error_pos_attr) == cfg_error_attr
          PutStrAttrXY(error_scr_col, error_scr_row, error_pos_char, saved_attr)
        else
          saved_attr = error_pos_attr
          PutStrAttrXY(error_scr_col, error_scr_row, error_pos_char, Chr(cfg_error_attr))
        endif
      endif
    endif
  endif
end show_error_cursor


proc after_getkey()
  if is_json_buffer()
    SetBufferInt(RECHECK_VAR_NAME, TRUE)
  endif
end after_getkey


proc idle()
  idle_timer = idle_timer + 1

  if idle_timer >= 5
    idle_timer = 0

    if GetBufferInt(RECHECK_VAR_NAME)
      SetBufferInt(RECHECK_VAR_NAME, FALSE)

      if FileChanged() <> GetBufferInt(FILECHANGED_VAR_NAME)
        SetBufferInt(FILECHANGED_VAR_NAME, FileChanged())
        check_for_json_syntax_errors(_HIDDEN_)
      endif
    endif

    if GetBufferInt(ERROR_LINE_VAR_NAME)
      show_error_cursor()
    endif
  endif
end idle


proc on_changing_files()
  if is_json_buffer()
    idle_timer       = 0
    prev_error_state = FALSE
    SetBufferInt(RECHECK_VAR_NAME    , TRUE)
    SetBufferInt(FILECHANGED_VAR_NAME, -1  )
  endif
end on_changing_files


proc on_file_save()
  integer choice    = FALSE
  integer save_file = FALSE

  restore_id = 0

  if  is_json_buffer()
  and check_for_json_syntax_errors(_NORMAL_)
    case cfg_save_option
      when 1, 2
        UpdateDisplay(_ALL_WINDOWS_REFRESH_)
        repeat
          if cfg_save_option == 1
            PushKey(<Tab>)
          endif

          to_beep_or_not_to_beep()
          choice = MsgBox(MACRO_NAME,
                          'Current .json buffer has errors. Save anyway?',
                          _YES_NO_)
        until choice
        save_file = iif(choice == 1, TRUE, FALSE)
      when 3
        UpdateDisplay(_ALL_WINDOWS_REFRESH_)
        pop_message(Format('File has errors and is not saved:';
                           SplitPath(CurrFilename(), _NAME_|_EXT_)))
        save_file = FALSE
      when 4
        UpdateDisplay(_ALL_WINDOWS_REFRESH_)
        pop_message(Format('File has errors and is saved anyway';
                           SplitPath(CurrFilename(), _NAME_|_EXT_)))
        save_file = TRUE
      when 5
        save_file = FALSE
      when 6
        save_file = TRUE
    endcase

    if not save_file
      restore_id      = GetBufferId()
      old_undomode    = UndoMode(ON)
      old_undocount   = UndoCount()
      old_fqn         = CurrFilename()
      ChangeCurrFilename(TMP_FQN, CCF_OPTIONS)
      push_position()

      // Bug: EmptyBuffer() clears undo information. Use work-around instead:
      PushBlock()
      MarkLine(1, NumLines())
      KillBlock()
      PopBlock()
    endif
  endif
end on_file_save


proc after_file_save()
  integer loop_counter = 0

  if restore_id
    ChangeCurrFilename(old_fqn, CCF_OPTIONS)

    //  In a test I counted 8 undo loops.
    //  Theoretically other macros could have made additional changes
    //  that now need to be undone too.
    while loop_counter < 10000
    and   UndoCount()  > old_undocount
      Undo()
      loop_counter = loop_counter + 1
    endwhile

    if loop_counter >= 10000
      Warn(MACRO_NAME;
           'program error: Infinite loop aborted. Beware other errors!')
    endif

    pop_position()
    UndoMode(old_undomode)
    restore_id = 0
  endif
end after_file_save


proc validate_current_buffer()
  integer ok = FALSE

  PushLocation()
  BegFile()

  ok = ParseJson()

  if ok
    PopPosition()
  else
    KillPosition()
  endif
end validate_current_buffer


string proc get_message_duration()
  return(Str(cfg_message_duration))
end get_message_duration


proc set_message_duration()
  string s[3] = Str(cfg_message_duration)

  cfg_message_duration = iif(ReadNumeric(s),
                             Max(Val(s), 1),
                             cfg_message_duration)
  write_profile_int(CFG_SECTION_NAME, 'MessageDuration', cfg_message_duration)
end set_message_duration


integer proc get_message_duration_flag()
  integer flag = _MF_GRAYED_|_MF_SKIP_

  if cfg_save_option in 3, 4
    flag = _MF_ENABLED_|_MF_DONT_CLOSE_
  endif

  return(flag)
end get_message_duration_flag


string proc get_save_option()
  return(GetToken(save_options, ';', cfg_save_option))
end get_save_option


proc set_save_option(integer save_option)
  cfg_save_option = save_option
  write_profile_int(CFG_SECTION_NAME, 'SaveOption', cfg_save_option)

  restart_main_menu = TRUE
  PushKey(<Escape>)
end set_save_option


menu configure_save_option()
  title   = 'JSON Validator Save Option'
  history = cfg_save_option

  'Ask ,default No',
    set_save_option(1),
    _MF_ENABLED_|_MF_CLOSE_BEFORE_,
    ''
  'Ask, default Yes,',
    set_save_option(2),
    _MF_ENABLED_|_MF_CLOSE_BEFORE_,
    ''
  'No , with temporary message',
    set_save_option(3),
    _MF_ENABLED_|_MF_CLOSE_BEFORE_,
    ''
  'Yes, with temporary message',
    set_save_option(4),
    _MF_ENABLED_|_MF_CLOSE_BEFORE_,
    ''
  'No , silently',
    set_save_option(5),
    _MF_ENABLED_|_MF_CLOSE_BEFORE_,
    ''
  'Yes, silently',
    set_save_option(6),
    _MF_ENABLED_|_MF_CLOSE_BEFORE_,
    ''
end configure_save_option


menu main_menu()
  title = 'JSON Validator'
  history

  '',, _MF_SKIP_
  'Action',, _MF_DIVIDE_
  '',, _MF_SKIP_

  '&Validate current buffer for JSON syntax ...',
    validate_current_buffer(),
    _MF_CLOSE_ALL_BEFORE_|_MF_ENABLED_,
    'Validate the current buffer for JSON syntax'

  '',, _MF_SKIP_
  'Configuration',, _MF_DIVIDE_
  '',, _MF_SKIP_

  '&Save if errors'
    [get_save_option():27],
    configure_save_option(),
    _MF_ENABLED_,
    'What to do when saving a .json file that has syntax errors?'

  '  Message duration (s)'
    [get_message_duration():3],
    set_message_duration(),
    get_message_duration_flag(),
    'If showing a temporary message, then for how many 18ths of a second?'

  '&Color for error cursor'
    [get_color_name(cfg_error_attr):MAX_COLOR_NAME_SIZE],
    configure_color('Error Cursor'),
    _MF_ENABLED_,
    'Select the color for the cursor that shows the location of a syntax error'

  '',, _MF_SKIP_
  'Help',, _MF_DIVIDE_
  '',, _MF_SKIP_

  '&Help ...',
    show_help(),,
    'Show the documentation'
  '',, _MF_SKIP_
end main_menu


proc WhenPurged()
end WhenPurged


proc WhenLoaded()
  MACRO_NAME           = SplitPath(CurrMacroFilename(), _NAME_)

  CFG_SECTION_NAME     = MACRO_NAME + ':Configuration'

  ERROR_LINE_VAR_NAME  = MACRO_NAME + ':ErrorLine'
  ERROR_POS_VAR_NAME   = MACRO_NAME + ':ErrorPos'
  FILECHANGED_VAR_NAME = MACRO_NAME + ':FileChanged'
  RECHECK_VAR_NAME     = MACRO_NAME + ':Recheck'

  TMP_FQN              = GetTempPath() + 'TSE_' + MACRO_NAME + '.tmp'

  cfg_error_attr       = GetProfileInt(CFG_SECTION_NAME,
                                       'ErrorColor',
                                       Color(Bright White ON Red))

  cfg_save_option      = GetProfileInt(CFG_SECTION_NAME,
                                       'SaveOption',
                                       1)

  cfg_message_duration = GetProfileInt(CFG_SECTION_NAME,
                                       'MessageDuration',
                                       5)

  Hook(_AFTER_FILE_SAVE_  , after_file_save  )
  Hook(_AFTER_GETKEY_     , after_getkey     )
  Hook(_IDLE_             , idle             )
  Hook(_ON_CHANGING_FILES_, on_changing_files)
  Hook(_ON_FILE_SAVE_     , on_file_save     )
end WhenLoaded


proc Main()
  if Lower(Query(MacroCmdLine)) == 'validate'
    ParseJson()
  else
    repeat
      restart_main_menu = FALSE
      main_menu()
    until not restart_main_menu
  endif
end Main

